#ifndef _LINUX_STRING_H_
#define _LINUX_STRING_H_

/* We don't want strings.h stuff being user by user stuff by accident */

#ifdef __KERNEL__

#include <linux/linkage.h>        /* for asmlinkage */
#include <linux/compiler.h>	/* for inline */
#include <linux/types.h>	/* for size_t */
#include <linux/stddef.h>	/* for NULL */

#ifdef __cplusplus
extern "C" {
#endif

extern char * NTS strpbrk(const char * NTS,const char * NTS);
extern char * NTS strsep(char * NTS *,const char * NTS);
extern __kernel_size_t strspn(const char * NTS,const char * NTS);
extern __kernel_size_t strcspn(const char * NTS,const char * NTS);

/*
 * Include machine specific inline routines
 */
#include <asm/string.h>

/* zf: these implementations are all from lib/string.c */
#ifndef __HAVE_ARCH_STRLEN
#ifdef DEPUTY
extern int deputy_strlen(const char * NTS str);
#define strlen(str) deputy_strlen(str)
#else
static __inline size_t strlen(const char * NTS s) {
	const char *sc;
	for (sc = s; *sc != '\0'; ++sc)
		/* nothing */;
	return sc - s;
}
#endif
#endif
#ifndef __HAVE_ARCH_STRCPY
/* zf: currently we cannot check this yet */
static __inline char * NTS strcpy(char * TRUSTED dest, const char * TRUSTED src)
{
	char * NTS tmp = dest;
	while ((*dest = *src) != '\0') {
            dest++;
            src++;
        }
	return tmp;
}

#endif
#ifndef __HAVE_ARCH_STRNCPY
/* extern char * strncpy(char *,const char *, __kernel_size_t); */
/*
static __inline char * strncpy(char * TRUSTED dest,const char * NTS src, __kernel_size_t count) {
	char *tmp = dest;

	while (count) {
		if ((*tmp = *src) != 0) src++;
		tmp++;
		count--;
	}
	return dest;
}
*/
extern char * NTS deputy_strncpy(char * TRUSTED, const char * NTS, __kernel_size_t);
#define strncpy(a,b,c) deputy_strncpy((a),(b),(c))
#endif
#ifndef __HAVE_ARCH_MEMCPY
extern void * (DMEMCPY(1, 2, 3) memcpy)(void *,const void *,__kernel_size_t);
#define __memcpy(x,y,z) memcpy(x,y,z)
#endif
#ifndef __HAVE_ARCH_STRLCPY
static __inline size_t strlcpy(char * COUNT(size-1) NULLTERM dest, const char * NTS src, size_t size)
{
	size_t ret = strlen(src);

	if (size) {
		size_t len = (ret >= size) ? size-1 : ret;
		memcpy(dest, src, len);
		dest[len] = '\0';
	}
	return ret;
}
#endif
#ifndef __HAVE_ARCH_STRCAT
static __inline char * NTS strcat(char * TRUSTED dest, const char * NTS src)
{
	char * NTS tmp = dest;
        const char * NTS src2 = src;

	while (*dest) {
		dest++;
        }

	while ((*dest = *src2) != '\0') {
            dest++;
            src2++;
        }

	return tmp;
}
#endif
#ifndef __HAVE_ARCH_STRNCAT
static __inline char * NTS strncat(char * TRUSTED dest, const char * NTS src, size_t count)
{
	char * NTS tmp = dest;
	const char * NTS src2 = src;

	if (count) {
		while (*dest) {
			dest++;
                }
		while ((*dest = *src2) != 0) {
                        dest++;
                        src2++;
			if (--count == 0) {
				*dest = '\0';
				break;
			}
		}
	}

	return tmp;
}
#endif
#ifndef __HAVE_ARCH_STRLCAT
static __inline size_t strlcat(char * NT COUNT(count-1) dest, const char * NTS src, __kernel_size_t count) {
	size_t dsize = strlen(dest);
	size_t len = strlen(src);
	size_t res = dsize + len;

	/* This would be a bug */
	BUG_ON(dsize >= count);

	dest += dsize;
	count -= dsize;
	if (len >= count)
		len = count-1;
	memcpy(dest, src, len);
	dest[len] = 0;
	return res;
}

#endif
#ifndef __HAVE_ARCH_STRCMP
extern int strcmp(const char * NTS,const char * NTS);
#endif
#ifndef __HAVE_ARCH_STRNCMP
static __inline int strncmp(const char * NTS cs, const char * NTS ct, size_t count)
{
	register signed char __res = 0;
	const char *cs2 = cs;
	const char *ct2 = ct;

	while (count) {
		if ((__res = *cs2 - *ct2++) != 0 || !*cs2++)
			break;
		count--;
	}

	return __res;
}
#endif
#ifndef __HAVE_ARCH_STRNICMP
extern int strnicmp(const char *, const char *, __kernel_size_t);
#endif
#ifndef __HAVE_ARCH_STRCHR
extern char * NTS strchr(const char * NTS,int);
#endif
#ifndef __HAVE_ARCH_STRNCHR
extern char * NTS strnchr(const char * NT COUNT(count-1), size_t count, int);
#endif
#ifndef __HAVE_ARCH_STRRCHR
extern char * NTS strrchr(const char * NTS,int);
#endif
#ifndef __HAVE_ARCH_STRSTR
extern char * NTS strstr(const char * NTS,const char * NTS);
#endif
#ifndef __HAVE_ARCH_STRNLEN
extern __kernel_size_t strnlen(const char * NT COUNT(count-1),__kernel_size_t count);
#endif

#ifndef __HAVE_ARCH_MEMSET
extern void * (DMEMSET(1, 2, 3) memset)(void * s,int c,__kernel_size_t n);
#endif
#ifndef __HAVE_ARCH_MEMMOVE
extern void * (DMEMCPY(1, 2, 3) memmove)(void *,const void *,__kernel_size_t);
#endif
#ifndef __HAVE_ARCH_MEMSCAN
extern void * memscan(void * SIZE(size),int,__kernel_size_t size);
#endif
#ifndef __HAVE_ARCH_MEMCMP
extern int (DMEMCMP(1, 2, 3) memcmp)(const void *,const void *,__kernel_size_t);
#endif
#ifndef __HAVE_ARCH_MEMCHR
extern void * memchr(const void * SIZE(n),int,__kernel_size_t n);
#endif

extern char * NTS kstrdup(const char * NTS s, gfp_t gfp);

#ifdef __cplusplus
}
#endif

#endif
#endif /* _LINUX_STRING_H_ */
